try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


import json
from dotenv import load_dotenv

from fastmcp import Client

from importlib.metadata import metadata

import httpx
from mcp.types import TextContent

import os
from typing import Annotated, Any

from pydantic import BaseModel, Field

SPACE_STRING = " "
TRUE_VALUES_LIST = ["true", "1", "yes", "on", "yes", "y", "t"]
FALSE_VALUES_LIST = ["false", "0", "no", "off", "n", "f"]


class EnvironmentVariables:
    """
    List of environment variables used in this project.
    """

    MCP_SERVER_HOST = "FASTMCP_HOST"
    DEFAULT__MCP_SERVER_HOST = "localhost"

    MCP_SERVER_PORT = "FASTMCP_PORT"
    DEFAULT__MCP_SERVER_PORT = 8000

    MCP_SERVER_TRANSPORT = "MCP_SERVER_TRANSPORT"
    DEFAULT__MCP_SERVER_TRANSPORT = "stdio"
    ALLOWED__MCP_SERVER_TRANSPORT = ["stdio", "sse", "streamable-http"]

    MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE = "MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE"
    DEFAULT__MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE = True

    FRANKFURTER_API_URL = "FRANKFURTER_API_URL"
    DEFAULT__FRANKFURTER_API_URL = "https://api.frankfurter.dev/v1"

    HTTPX_TIMEOUT = "HTTPX_TIMEOUT"
    DEFAULT__HTTPX_TIMEOUT = 5.0

    HTTPX_VERIFY_SSL = "HTTPX_VERIFY_SSL"
    DEFAULT__HTTPX_VERIFY_SSL = True


class AppMetadata:
    """
    Metadata for the application.
    """

    PACKAGE_NAME = "frankfurtermcp"

    TEXT_CONTENT_META_PREFIX = "frankfurtermcp."


package_metadata = metadata(AppMetadata.PACKAGE_NAME)


class ResponseMetadata(BaseModel):
    """
    Metadata for the TextContent response.
    """

    version: Annotated[str, Field(description="The version of the package.")]
    api_url: Annotated[
        str,
        Field(description="The URL of the Frankfurter API used for the call."),
    ]
    api_status_code: Annotated[
        int, Field(description="The HTTP status code of the API response.")
    ]
    api_bytes_downloaded: Annotated[
        int, Field(description="The number of bytes downloaded in the API response.")
    ]
    api_elapsed_time: Annotated[
        int, Field(description="The elapsed time for the API call in microseconds.")
    ]


class CurrencyConversionResponse(BaseModel):
    """
    Response model for currency conversion.
    """

    from_currency: Annotated[str, Field(description="The currency to convert from.")]
    to_currency: Annotated[str, Field(description="The currency to convert to.")]
    amount: Annotated[
        float, Field(description="The amount (of the source currency) to convert.")
    ]
    converted_amount: Annotated[
        float, Field(description="The converted amount (of the target currency).")
    ]
    exchange_rate: Annotated[
        float, Field(description="The exchange rate used for the conversion.")
    ]
    rate_date: Annotated[
        str,
        Field(
            description="The date, ISO format, of the exchange rate used for the conversion."
        ),
    ]


load_dotenv()


def parse_env(
    var_name: str,
    default_value: Any | None = None,
    allowed_values: list[Any] | None = None,
    type_cast: type = str,
    convert_to_list=False,
    list_split_char=SPACE_STRING,
) -> Any | list[Any]:
    """
    Parse the environment variable and return the value.

    Args:
        var_name (str): The name of the environment variable.
        default_value (Any | None): The default value to use if the environment variable is not set. Defaults to None.
        allowed_values (list[Any] | None): A list of allowed values for the environment variable. If provided, the
            value will be checked against this list. This option is ignored if type_cast is bool.
        type_cast (type): The type to cast the value to.
        convert_to_list (bool): Whether to convert the value to a list.
        list_split_char (str): The character to split the list on.

    Returns:
        (Any | list[Any]) The parsed value, either as a single value or a list. The type of the returned single
        value or individual elements in the list depends on the supplied type_cast parameter.
    """
    if default_value is not None and not isinstance(default_value, type_cast):
        raise TypeError(
            f"The default value {default_value} specified for the environment variable {var_name} is of type {type(default_value).__name__}. However, the expected type is {type_cast.__name__} instead."
        )
    if os.getenv(var_name) is None and default_value is None:
        raise ValueError(
            f"Environment variable {var_name} does not exist and a default value has not been provided."
        )
    parsed_value = None
    if type_cast is bool:
        # Sometimes, the environment variable is set to a string that represents a boolean value.
        # We convert it to lowercase and check against TRUE_VALUES_LIST.
        # The following logic also works if the boolean value is set to a True or False boolean type.
        parsed_value = (
            str(os.getenv(var_name, default_value)).lower() in TRUE_VALUES_LIST
        )
    else:
        parsed_value = os.getenv(var_name, default_value)
        if allowed_values is not None:
            if parsed_value not in allowed_values:
                raise ValueError(
                    f"Environment variable {var_name} has value '{parsed_value}', "
                    f"which is not in the allowed values: {allowed_values}."
                )

    if not convert_to_list:
        value: Any = (
            type_cast(parsed_value)
            if not isinstance(parsed_value, type_cast)
            else parsed_value
        )
    else:
        value: list[Any] = [
            (type_cast(v) if not isinstance(v, type_cast) else v)
            for v in parsed_value.split(list_split_char)
        ]
    return value


frankfurter_api_url = parse_env(
    EnvironmentVariables.FRANKFURTER_API_URL,
    default_value=EnvironmentVariables.DEFAULT__FRANKFURTER_API_URL,
)


def get_nonstdio_transport_endpoint() -> str:
    """
    Get the transport endpoint for the non-stdio MCP client based on environment variables.
    """
    transport_type = parse_env(
        EnvironmentVariables.MCP_SERVER_TRANSPORT,
        default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_TRANSPORT,
        allowed_values=EnvironmentVariables.ALLOWED__MCP_SERVER_TRANSPORT,
    )
    transport_endpoint = None
    server_host = parse_env(
        "FASTMCP_HOST",
        default_value="localhost",
    )
    server_port = parse_env(
        "FASTMCP_PORT",
        default_value=8000,
        type_cast=int,
    )
    if transport_type == "streamable-http":
        transport_endpoint = f"http://{server_host}:{server_port}/mcp"
    elif transport_type == "sse":
        transport_endpoint = f"http://{server_host}:{server_port}/sse"
    else:
        raise ValueError(
            f"Unsupported transport type: {transport_type}. "
            "Allowed values are for this type of client are only sse and streamable-http: "
        )
    return transport_endpoint


def get_nonstdio_mcp_client() -> Client:
    """
    Create and return a non-stdio MCP client to connect to built-in server based on the environment variables.
    """
    transport_endpoint = get_nonstdio_transport_endpoint()
    mcp_client = Client(
        transport=transport_endpoint,
        timeout=60,
    )
    return mcp_client


def get_text_content(
    data: Any,
    http_response: httpx.Response,
    include_metadata: bool = parse_env(
        var_name=EnvironmentVariables.MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE,
        default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE,
        type_cast=bool,
    ),
) -> TextContent:
    """
    Convert data to TextContent format.

    Args:
        data (Any): The data to convert.
        include_metadata (bool): Whether to include metadata in the TextContent.

    Returns:
        TextContent: The converted TextContent object.
    """
    literal_text = "text"
    if isinstance(data, TextContent):
        # do nothing yet
        pass
    elif isinstance(data, (str, int, float, complex, bool, type(None))):
        text_content = TextContent(type=literal_text, text=str(data))
    elif isinstance(data, dict) or isinstance(data, list):
        text_content = TextContent(type=literal_text, text=json.dumps(data))
    elif isinstance(data, BaseModel):
        text_content = TextContent(type=literal_text, text=data.model_dump_json())
    else:
        raise TypeError(
            f"Unsupported data type: {type(data).__name__}. "
            "Only str, int, float, complex, bool, dict, list, and Pydantic BaseModel types are supported for wrapping as TextContent."
        )
    if include_metadata:
        text_content.meta = text_content.meta if hasattr(text_content, "meta") else {}
        text_content.meta[AppMetadata.PACKAGE_NAME] = ResponseMetadata(
            version=package_metadata["Version"],
            api_url=frankfurter_api_url,
            api_status_code=http_response.status_code,
            api_bytes_downloaded=http_response.num_bytes_downloaded,
            api_elapsed_time=http_response.elapsed.microseconds,
        ).model_dump()
    return text_content



from datetime import date
import os
import signal
import sys
from typing import Annotated, List, Optional
import certifi
import httpx
import ssl
from dotenv import load_dotenv

from fastmcp import FastMCP, Context

from pydantic import Field
from rich import print as print
from frankfurtermcp.common import (
    CurrencyConversionResponse,
    EnvironmentVariables,
    get_text_content,
    parse_env,
)

from frankfurtermcp.common import package_metadata, frankfurter_api_url

app = FastMCP(
    name=package_metadata["Name"],
    instructions=package_metadata["Description"],
    on_duplicate_prompts="error",
    on_duplicate_resources="error",
    on_duplicate_tools="error",
)


def _obtain_httpx_client() -> httpx.Client:
    """
    Obtain an HTTPX client for making requests to the Frankfurter API.
    """
    verify = parse_env(
        EnvironmentVariables.HTTPX_VERIFY_SSL,
        default_value=EnvironmentVariables.DEFAULT__HTTPX_VERIFY_SSL,
        type_cast=bool,
    )
    if verify is False:
        print(
            "[yellow]SSL verification is disabled. This is not recommended for production use.[/yellow]"
        )
    ctx = ssl.create_default_context(
        cafile=os.environ.get("SSL_CERT_FILE", certifi.where()),
        capath=os.environ.get("SSL_CERT_DIR"),
    )
    client = httpx.Client(
        verify=verify if (verify is not None and verify is False) else ctx,
        follow_redirects=True,
        trust_env=True,
        timeout=parse_env(
            EnvironmentVariables.HTTPX_TIMEOUT,
            default_value=EnvironmentVariables.DEFAULT__HTTPX_TIMEOUT,
            type_cast=float,
        ),
    )
    return client


@app.tool(
    tags=["currency-rates", "supported-currencies"],
    annotations={
        "readOnlyHint": True,
        "openWorldHint": True,
    },
)
async def get_supported_currencies(ctx: Context) -> list[dict]:
    """
    Returns a list of three-letter currency codes for the supported currencies.
    """
    try:
        with _obtain_httpx_client() as client:
            await ctx.info(
                f"Fetching supported currencies from Frankfurter API at {frankfurter_api_url}"
            )
            http_response = client.get(f"{frankfurter_api_url}/currencies")
            result = http_response.json()
            client.close()
            return get_text_content(data=result, http_response=http_response)
    except httpx.RequestError as e:
        raise ValueError(
            f"Failed to fetch supported currencies from {frankfurter_api_url}: {e}"
        )
    except ValueError as e:
        raise ValueError(f"Failed to parse response from {frankfurter_api_url}: {e}")


def _get_latest_exchange_rates(base_currency: str = None, symbols: list[str] = None):
    """
    Internal function to get the latest exchange rates.
    This is a helper function for the main tool.
    """
    try:
        params = {}
        if base_currency:
            params["base"] = base_currency
        if symbols:
            params["symbols"] = ",".join(symbols)
        with _obtain_httpx_client() as client:
            http_response = client.get(
                f"{frankfurter_api_url}/latest",
                params=params,
            )
            result = http_response.json()
            client.close()
            return result, http_response
    except httpx.RequestError as e:
        raise ValueError(
            f"Failed to fetch latest exchange rates from {frankfurter_api_url}: {e}"
        )
    except ValueError as e:
        raise ValueError(f"Failed to parse response from {frankfurter_api_url}: {e}")


def _get_historical_exchange_rates(
    specific_date: str = None,
    start_date: str = None,
    end_date: str = None,
    base_currency: str = None,
    symbols: list[str] = None,
) -> dict:
    """
    Internal function to get historical exchange rates.
    This is a helper function for the main tool.
    """
    try:
        params = {}
        if base_currency:
            params["base"] = base_currency
        if symbols:
            params["symbols"] = ",".join(symbols)

        frankfurter_url = frankfurter_api_url
        if start_date and end_date:
            frankfurter_url += f"/{start_date}..{end_date}"
        elif start_date:
            # If only start_date is provided, we assume the end date is the latest available date
            frankfurter_url += f"/{start_date}.."
        elif specific_date:
            # If only specific_date is provided, we assume it is the date for which we want the rates
            frankfurter_url += f"/{specific_date}"
        else:
            raise ValueError(
                "You must provide either a specific date, a start date, or a date range."
            )

        with _obtain_httpx_client() as client:
            http_response = client.get(
                frankfurter_url,
                params=params,
            )
            result = http_response.json()
            client.close()
            return result, http_response
    except httpx.RequestError as e:
        raise ValueError(
            f"Failed to fetch historical exchange rates from {frankfurter_api_url}: {e}"
        )
    except ValueError as e:
        raise ValueError(f"Failed to parse response from {frankfurter_api_url}: {e}")


@app.tool(
    tags=["currency-rates", "exchange-rates"],
    annotations={
        "readOnlyHint": True,
        "openWorldHint": True,
    },
)
async def get_latest_exchange_rates(
    ctx: Context,
    base_currency: Annotated[
        str,
        Field(description="A base currency code for which rates are to be requested."),
    ],
    symbols: Annotated[
        Optional[List[str]],
        Field(
            description="A list of target currency codes for which rates against the base currency will be provided. Do not provide it to request all supported currencies."
        ),
    ] = None,
) -> dict:
    """
    Returns the latest exchange rates for specific currencies. The
    symbols can be used to filter the results to specific currencies.
    If symbols is not provided, all supported currencies will be returned.
    """
    await ctx.info(
        f"Fetching latest exchange rates from Frankfurter API at {frankfurter_api_url}"
    )
    result, http_response = _get_latest_exchange_rates(
        base_currency=base_currency,
        symbols=symbols,
    )
    return get_text_content(data=result, http_response=http_response)


@app.tool(
    tags=["currency-rates", "currency-conversion"],
    annotations={
        "readOnlyHint": True,
        "openWorldHint": True,
    },
)
async def convert_currency_latest(
    ctx: Context,
    amount: Annotated[
        float, Field(description="The amount in the source currency to convert.")
    ],
    from_currency: Annotated[str, Field(description="The source currency code.")],
    to_currency: Annotated[str, Field(description="The target currency code.")],
) -> dict:
    """
    Converts an amount from one currency to another using the latest exchange rates.
    """
    if from_currency.lower() == to_currency.lower():
        # If the source and target currencies are the same, no conversion is needed
        raise ValueError(
            f"Source currency '{from_currency}' and target currency '{to_currency}' are the same. No conversion needed."
        )
    await ctx.info(
        f"Obtaining latest exchange rates for {from_currency} to {to_currency} from Frankfurter API at {frankfurter_api_url}"
    )
    latest_rates, http_response = _get_latest_exchange_rates(
        base_currency=from_currency,
        symbols=[to_currency],
    )
    await ctx.info(f"Converting {amount} of {from_currency} to {to_currency}")
    if not latest_rates or "rates" not in latest_rates:
        raise ValueError(
            f"Could not retrieve exchange rates for {from_currency} to {to_currency}."
        )
    rate = latest_rates["rates"].get(to_currency)
    if rate is None:
        raise ValueError(
            f"Exchange rate for {from_currency} to {to_currency} not found."
        )
    converted_amount = amount * float(rate)
    result = CurrencyConversionResponse(
        from_currency=from_currency,
        to_currency=to_currency,
        amount=amount,
        converted_amount=converted_amount,
        exchange_rate=rate,
        rate_date=latest_rates["date"],
    )
    return get_text_content(data=result, http_response=http_response)


@app.tool(
    tags=["currency-rates", "historical-exchange-rates"],
    annotations={
        "readOnlyHint": True,
        "openWorldHint": True,
    },
)
async def get_historical_exchange_rates(
    ctx: Context,
    base_currency: Annotated[
        str,
        Field(description="A base currency code for which rates are to be requested."),
    ],
    symbols: Annotated[
        Optional[List[str]],
        Field(
            description="A list of target currency codes for which rates against the base currency will be provided. Do not provide it to request all supported currencies."
        ),
    ] = None,
    specific_date: Annotated[
        Optional[date],
        Field(
            default=None,
            description="The specific date for which the historical rates are requested in the YYYY-MM-DD format.",
        ),
    ] = None,
    start_date: Annotated[
        Optional[date],
        Field(
            default=None,
            description="The start date, of a date range, for which the historical rates are requested in the YYYY-MM-DD format.",
        ),
    ] = None,
    end_date: Annotated[
        Optional[date],
        Field(
            default=None,
            description="The end date, of a date range, for which the historical rates are requested in the YYYY-MM-DD format.",
        ),
    ] = None,
) -> dict:
    """
    Returns historical exchange rates for a specific date or date range.
    If the exchange rates for a specified date is not available, the rates available for
    the closest date before the specified date will be provided.
    Either a specific date, a start date, or a date range must be provided.
    The symbols can be used to filter the results to specific currencies.
    If symbols are not provided, all supported currencies will be returned.
    """
    await ctx.info(
        f"Fetching historical exchange rates from Frankfurter API at {frankfurter_api_url}"
    )
    result, http_response = _get_historical_exchange_rates(
        specific_date=specific_date,
        start_date=start_date,
        end_date=end_date,
        base_currency=base_currency,
        symbols=symbols,
    )
    await ctx.info(
        f"Historical exchange rates fetched for {len(result.get('rates', []))} dates."
    )
    return get_text_content(data=result, http_response=http_response)


@app.tool(
    tags=["currency-rates", "currency-conversion", "historical-exchange-rates"],
    annotations={
        "readOnlyHint": True,
        "openWorldHint": True,
    },
)
async def convert_currency_specific_date(
    ctx: Context,
    amount: Annotated[
        float, Field(description="The amount in the source currency to convert.")
    ],
    from_currency: Annotated[str, Field(description="The source currency code.")],
    to_currency: Annotated[str, Field(description="The target currency code.")],
    specific_date: Annotated[
        date,
        Field(
            description="The specific date for which the conversion is requested in the YYYY-MM-DD format."
        ),
    ],
) -> dict:
    """
    Convert an amount from one currency to another using the exchange rates for a specific date.
    If there is no exchange rate available for the specific date, the rate for the closest available date before
    the specified date will be used.
    """
    if from_currency.lower() == to_currency.lower():
        # If the source and target currencies are the same, no conversion is needed
        raise ValueError(
            f"Source currency '{from_currency}' and target currency '{to_currency}' are the same. No conversion needed."
        )
    await ctx.info(
        f"Obtaining historical exchange rates for {from_currency} to {to_currency} on {specific_date} from Frankfurter API at {frankfurter_api_url}"
    )
    date_specific_rates, http_response = _get_historical_exchange_rates(
        specific_date=specific_date,
        base_currency=from_currency,
        symbols=[to_currency],
    )
    await ctx.info(
        f"Converting {amount} of {from_currency} to {to_currency} on {specific_date}"
    )
    if not date_specific_rates or "rates" not in date_specific_rates:
        raise ValueError(
            f"Could not retrieve exchange rates for {from_currency} to {to_currency} for {specific_date}."
        )
    rate = date_specific_rates["rates"].get(to_currency)
    if rate is None:
        raise ValueError(
            f"Exchange rate for {from_currency} to {to_currency} not found."
        )
    converted_amount = amount * float(rate)
    result = CurrencyConversionResponse(
        from_currency=from_currency,
        to_currency=to_currency,
        amount=amount,
        converted_amount=converted_amount,
        exchange_rate=rate,
        rate_date=date_specific_rates["date"],
    )
    return get_text_content(data=result, http_response=http_response)


def main():
    def sigint_handler(signal, frame):
        """
        Signal handler to shut down the server gracefully.
        """
        # Is this handler necessary since we are not doing anything and uvicorn already handles this?
        # Commenting out the following status outputs because it may confuse some MCP clients.
        # print("[green]Attempting graceful shutdown[/green], please wait...")
        # This is absolutely necessary to exit the program
        sys.exit(0)

    load_dotenv()
    # TODO: Should we also catch SIGTERM, SIGKILL, etc.? What about Windows?
    signal.signal(signal.SIGINT, sigint_handler)

    # Commenting out the following status outputs because it may confuse some MCP clients.
    # print(
    #     f"[green]Initiating startup[/green] of [bold]{package_metadata['Name']} {package_metadata['Version']}[/bold], [red]press CTRL+C to exit...[/red]"
    # )
    transport_type = parse_env(
        EnvironmentVariables.MCP_SERVER_TRANSPORT,
        default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_TRANSPORT,
        allowed_values=EnvironmentVariables.ALLOWED__MCP_SERVER_TRANSPORT,
    )
    (
        app.run(
            transport=transport_type,
            host=parse_env(
                EnvironmentVariables.MCP_SERVER_HOST,
                default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_HOST,
            ),
            port=parse_env(
                EnvironmentVariables.MCP_SERVER_PORT,
                default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_PORT,
                type_cast=int,
            ),
            uvicorn_config={
                "timeout_graceful_shutdown": 5,  # seconds
            },
        )
        if transport_type != "stdio"
        else app.run(transport=transport_type)
    )


if __name__ == "__main__":
    main()



import signal
import sys
from fastmcp import FastMCP
from frankfurtermcp.common import EnvironmentVariables
from frankfurtermcp.server import app as frankfurtermcp
from frankfurtermcp.common import parse_env
from dotenv import load_dotenv

app = FastMCP(
    name="test_composition",
    instructions="This is a MCP server to test dynamic composition of MCP.",
)

COMPOSITION_PREFIX = "composition_"


@app.tool(
    description="The quintessential hello world tool",
    tags=["hello", "world"],
    name="hello_world",
    annotations={
        "readOnlyHint": True,
    },
)
def hello_world(name: str = None) -> str:
    """
    A simple tool that returns a greeting message.

    Args:
        name (str): The name to greet.

    Returns:
        str: A greeting message.
    """
    suffix = "This is the MCP server to test dynamic composition."
    return f"Hello, {name}! {suffix}" if name else f"Hello World! {suffix}"


def main():
    """
    Main function to run the MCP server.
    """

    def sigint_handler(signal, frame):
        """
        Signal handler to shut down the server gracefully.
        """
        # This is absolutely necessary to exit the program
        sys.exit(0)

    load_dotenv()
    signal.signal(signal.SIGINT, sigint_handler)

    app.mount(prefix=COMPOSITION_PREFIX, server=frankfurtermcp, as_proxy=False)
    transport_type = parse_env(
        EnvironmentVariables.MCP_SERVER_TRANSPORT,
        default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_TRANSPORT,
        allowed_values=EnvironmentVariables.ALLOWED__MCP_SERVER_TRANSPORT,
    )
    (
        app.run(
            transport=transport_type,
            host=parse_env(
                EnvironmentVariables.MCP_SERVER_HOST,
                default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_HOST,
            ),
            port=parse_env(
                EnvironmentVariables.MCP_SERVER_PORT,
                default_value=EnvironmentVariables.DEFAULT__MCP_SERVER_PORT,
                type_cast=int,
            ),
            uvicorn_config={
                "timeout_graceful_shutdown": 5,  # seconds
            },
        )
        if transport_type != "stdio"
        else app.run(transport=transport_type)
    )


if __name__ == "__main__":
    main()



import multiprocessing
import asyncio
import time
from fastmcp import Client

from frankfurtermcp.common import get_nonstdio_mcp_client

# from frankfurtermcp.common import EnvironmentVariables
from frankfurtermcp.composition import (
    COMPOSITION_PREFIX,
    main as mcp_server_composition,
)

import pytest


@pytest.fixture(scope="module")
def mcp_client():
    """
    Fixture to create a client for the MCP server.
    """
    return get_nonstdio_mcp_client()


@pytest.fixture(scope="module", autouse=True)
def mcp_server():
    print("[fixture] Starting MCP server process")
    proc = multiprocessing.Process(
        target=mcp_server_composition,
    )
    proc.start()

    async def wait_for_server(timeout: float = 5.0) -> bool:
        """
        Wait for the MCP server to start and be ready to accept connections.
        """
        client = get_nonstdio_mcp_client()
        ping_result = False
        start = time.time()
        while (time.time() - start) < timeout and not ping_result:
            try:
                async with client:
                    ping_result = await client.ping()
            except Exception:
                # Don't panic, give it a bit more time!
                time.sleep(0.1)
        return ping_result

    try:
        # Check connection to the server
        if not asyncio.run(wait_for_server()):
            raise RuntimeError(
                "The MCP server process did not start in time or is not responding."
            )
        yield  # Run tests now
    finally:
        print("[fixture] Stopping MCP server process")
        if proc.is_alive():
            proc.terminate()
            proc.join()


class TestMCPComposition:
    async def list_tools(self, mcp_client: Client):
        """
        Helper method to call a tool on the MCP server.
        """
        async with mcp_client:
            result = await mcp_client.list_tools()
            await mcp_client.close()
        return result

    def test_list_tools(self, mcp_client):
        """
        Test the whether we can list the tools starting with the Frankfurter MCP tool name prefix.
        """
        tools = asyncio.run(
            self.list_tools(
                mcp_client=mcp_client,
            )
        )
        for tool in tools:
            print(f"Tool name: {tool.name}. Description: {tool.description}")
        assert isinstance(tools, list), "Expected a list of tools"
        assert len(tools) == 6, "Expected 6 tools to be available"
        composed_tools = [
            tool for tool in tools if tool.name.startswith(COMPOSITION_PREFIX)
        ]
        assert all(
            (isinstance(tool.name, str) and tool.name.startswith(COMPOSITION_PREFIX))
            for tool in composed_tools
        ), f"All composed tool names start with '{COMPOSITION_PREFIX}'"
